// noinspection JSUnusedGlobalSymbols,JSUnusedLocalSymbols,UnnecessaryLocalVariableJS

/**
 * @param {ScriptRuntime} scriptRuntime
 * @param {org.mozilla.javascript.Scriptable | global} scope
 * @return {Internal.Mime}
 */
module.exports = function (scriptRuntime, scope) {

    let _ = {
        Mime: (/* @IIFE */ () => {
            /**
             * @param {string} mimeStr
             * @return {Internal.JsMime}
             */
            const __ = (mimeStr) => {
                return new _.JsMime(mimeStr);
            };

            const Mime = function () {
                return Object.setPrototypeOf(Object.assign(__, Mime.prototype), scriptRuntime.mime);
            };

            /* Reserved for extension. */
            Mime.prototype = {
                constructor: Mime,
            };

            return Mime;
        })(),
        JsMime: (() => {
            /**
             * @param {string} mimeStr
             * @extends {Internal.JsMime}
             */
            const JsMime = function (mimeStr) {
                let base = (/* @IIFE */ () => {
                    if (typeof mimeStr === 'string') {
                        let { type, subtype } = new eu.medsea.mimeutil.MimeType(mimeStr);
                        return { type, subtype };
                    }
                    throw TypeError(`MIME type "${mimeStr}" is invalid`);
                })();
                let { type, subtype } = base;
                let { mimeType, mimeTypeRefined, parameters } = _.parseFullMimeType(type, subtype, mimeStr);
                Object.assign(this, { raw: mimeStr, type, subtype, mimeType, mimeTypeRefined, parameters });
            };

            JsMime.prototype = {
                constructor: JsMime,
                toString() {
                    return `JsMime {\n` +
                        `  type: '${this.type}',\n` +
                        `  subtype: '${this.subtype}',\n` +
                        `  mimeType: '${this.mimeType}',\n` +
                        `  mimeTypeRefined: '${this.mimeTypeRefined}',\n` +
                        `  parameters: ${util.format(this.parameters)},\n` +
                        `  raw: '${this.raw}',\n` +
                        `}`;
                },
            };

            return JsMime;
        })(),
        parseFullMimeType(type, subtype, mimeType) {
            let paramObj = {};
            let mimeRefined = `${type}/${subtype}`;
            let mimeFull = mimeRefined;
            if (typeof mimeType === 'string') {
                let paramStrList = mimeType
                    .split(/\s*;\s*/)
                    .slice(1)
                    .filter(s => s.length > 0);
                if (paramStrList.length > 0) {
                    mimeFull += `; ${(paramStrList.join('; '))}`;
                    paramStrList.forEach((s) => {
                        if (s.includes('=')) {
                            let [ k, v ] = s.split(/\s*=\s*/);
                            paramObj[k] = v;
                        }
                    });
                }
            }
            return { mimeType: mimeFull, mimeTypeRefined: mimeRefined, parameters: paramObj };
        },
    };

    /**
     * @type {Internal.Mime}
     */
    const mime = new _.Mime();

    return mime;
};