// noinspection UnnecessaryLocalVariableJS,JSUnusedLocalSymbols

/**
 * @param {ScriptRuntime} scriptRuntime
 * @param {org.mozilla.javascript.Scriptable | global} scope
 * @return {Internal.Plugins}
 */
module.exports = function (scriptRuntime, scope) {
    let _ = {
        Plugins: (/* @IIFE */ () => {
            /**
             * @implements Internal.Plugins
             */
            const Plugins = function () {
                /* Empty body. */
            };

            Plugins.prototype = {
                constructor: Plugins,
                load(name) {
                    if (typeof name !== 'string') {
                        throw TypeError('The "name" argument for plugins.load() must be of type string');
                    }
                    if (name.includes('.') && !name.endsWith('.js')) /* As package name. */ {
                        let plugin = scriptRuntime.plugins.load(name);
                        let moduleExportedFunc = require(plugin.getMainScriptPath());
                        return moduleExportedFunc(plugin.unwrap());
                    }
                    if (files.exists('./plugins')) /* As project-level plugins name. */ {
                        return require(`./plugins/${name}`);
                    }
                    throw Error('A directory named "plugins" must be found in the root directory of current project');
                },
            };

            return Plugins;
        })(),
    };

    /**
     * @type {Internal.Plugins}
     */
    const plugins = new _.Plugins();

    return plugins;
};