/**
 * @param {ScriptRuntime} scriptRuntime
 * @param {org.mozilla.javascript.Scriptable | global} scope
 * @return {Internal.Shell}
 */
module.exports = function (scriptRuntime, scope) {
    let _ = {
        ShellCtor: (/* @IIFE */ () => {
            /**
             * @implements Internal.Shell
             */
            const ShellCtor = function () {
                /** @global */
                const shell = function (cmd, root) {
                    return scriptRuntime.shell(cmd, Number(Boolean(root)));
                };
                return Object.assign(shell, ShellCtor.prototype);
            };

            ShellCtor.prototype = {
                constructor: ShellCtor,
                fromIntent(i) {
                    return app.intentToShell(i);
                },
            };

            return ShellCtor;
        })(),
        scopeAugment() {
            Object.assign(scope, {
                /** @global */
                Menu: () => KeyCode(KeyEvent.KEYCODE_MENU),
                /** @global */
                Home: () => KeyCode(KeyEvent.KEYCODE_HOME),
                /** @global */
                Back: () => KeyCode(KeyEvent.KEYCODE_BACK),
                /** @global */
                Up: () => KeyCode(KeyEvent.KEYCODE_DPAD_UP),
                /** @global */
                Down: () => KeyCode(KeyEvent.KEYCODE_DPAD_DOWN),
                /** @global */
                Left: () => KeyCode(KeyEvent.KEYCODE_DPAD_LEFT),
                /** @global */
                Right: () => KeyCode(KeyEvent.KEYCODE_DPAD_RIGHT),
                /** @global */
                OK: () => KeyCode(KeyEvent.KEYCODE_DPAD_CENTER),
                /** @global */
                VolumeUp: () => KeyCode(KeyEvent.KEYCODE_VOLUME_UP),
                /** @global */
                VolumeDown: () => KeyCode(KeyEvent.KEYCODE_VOLUME_DOWN),
                /** @global */
                Power: () => KeyCode(KeyEvent.KEYCODE_POWER),
                /** @global */
                Camera: () => KeyCode(KeyEvent.KEYCODE_CAMERA),
                /** @global */
                Text: text => scriptRuntime.getRootShell().Text(text),
                /** @global */
                Input: text => scriptRuntime.getRootShell().Text(text),
                /** @global */
                Tap: (x, y) => scriptRuntime.getRootShell().Tap(x, y),
                /** @global */
                Screencap: path => scriptRuntime.getRootShell().Screencap(path),
                /** @global */
                KeyCode: keyCode => scriptRuntime.getRootShell().KeyCode(keyCode),
                /** @global */
                SetScreenMetrics: (w, h) => scriptRuntime.getRootShell().SetScreenMetrics(w, h),
                /** @global */
                Swipe: (x1, y1, x2, y2, duration) => duration === undefined
                    ? scriptRuntime.getRootShell().Swipe(x1, y1, x2, y2)
                    : scriptRuntime.getRootShell().Swipe(x1, y1, x2, y2, duration),
            });
        },
    };

    const shell = new _.ShellCtor();

    _.scopeAugment();

    return shell;
};